#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
import binascii
from pyvoltha.adapters.common.frameio.frameio import hexify
from twisted.python.failure import Failure
from unittest import TestCase, main, skip
from .mock.mock_adapter_agent import MockAdapterAgent
from .mock.mock_onu_handler import MockOnuHandler
from .mock.mock_olt_handler import MockOltHandler
from .mock.mock_onu import MockOnu
from pyvoltha.adapters.extensions.omci.omci_defs import *
from pyvoltha.adapters.extensions.omci.omci_frame import *
from pyvoltha.adapters.extensions.omci.omci_entities import *
from pyvoltha.adapters.extensions.omci.omci_me import ExtendedVlanTaggingOperationConfigurationDataFrame
from pyvoltha.adapters.extensions.omci.omci_cc import OMCI_CC, UNKNOWN_CLASS_ATTRIBUTE_KEY,\
    MAX_OMCI_REQUEST_AGE
from six.moves import range

DEFAULT_OLT_DEVICE_ID = 'default_olt_mock'
DEFAULT_ONU_DEVICE_ID = 'default_onu_mock'
DEFAULT_PON_ID = 0
DEFAULT_ONU_ID = 0
DEFAULT_ONU_SN = 'TEST00000001'

OP = EntityOperations
RC = ReasonCodes

successful = False
error_reason = None


def chunk(indexable, chunk_size):
    for i in range(0, len(indexable), chunk_size):
        yield indexable[i:i + chunk_size]


def hex2raw(hex_string):
    return ''.join(chr(int(byte, 16)) for byte in chunk(hex_string, 2))


class TestOmciCc(TestCase):
    """
    Test the Open OMCI Communication channels

    Note also added some testing of MockOnu behaviour since its behaviour during more
    complicated unit/integration tests may be performed in the future.
    """
    def setUp(self, let_msg_timeout=False):
        self.adapter_agent = MockAdapterAgent()

    def tearDown(self):
        if self.adapter_agent is not None:
            self.adapter_agent.tearDown()

    def setup_mock_olt(self, device_id=DEFAULT_OLT_DEVICE_ID):
        handler = MockOltHandler(self.adapter_agent, device_id)
        self.adapter_agent.add_device(handler.device)
        return handler

    def setup_mock_onu(self, parent_id=DEFAULT_OLT_DEVICE_ID,
                       device_id=DEFAULT_ONU_DEVICE_ID,
                       pon_id=DEFAULT_PON_ID,
                       onu_id=DEFAULT_ONU_ID,
                       serial_no=DEFAULT_ONU_SN):
        handler = MockOnuHandler(self.adapter_agent, parent_id, device_id, pon_id, onu_id)
        handler.serial_number = serial_no
        onu = MockOnu(serial_no, self.adapter_agent, handler.device_id) \
            if serial_no is not None else None
        handler.onu_mock = onu
        return handler

    def setup_one_of_each(self, timeout_messages=False):
        # Most tests will use at lease one or more OLT and ONU
        self.olt_handler = self.setup_mock_olt()
        self.onu_handler = self.setup_mock_onu(parent_id=self.olt_handler.device_id)
        self.onu_device = self.onu_handler.onu_mock
        self.adapter_agent.timeout_the_message = timeout_messages

        self.adapter_agent.add_child_device(self.olt_handler.device,
                                            self.onu_handler.device)

    def _is_omci_frame(self, results, omci_msg_type):
        assert isinstance(results, OmciFrame), 'Not OMCI Frame'
        assert 'omci_message' in results.fields, 'Not OMCI Frame'
        if omci_msg_type is not None:
            assert isinstance(results.fields['omci_message'], omci_msg_type)
        return results

    def _check_status(self, results, value):
        if value is not None: assert results is not None, 'unexpected emtpy message'
        status = results.fields['omci_message'].fields['success_code']
        assert status == value,\
            'Unexpected Status Code. Got {}, Expected: {}'.format(status, value)
        return results

    def _check_mib_sync(self, results, value):
        assert self.onu_device.mib_data_sync == value, \
            'Unexpected MIB DATA Sync value. Got {}, Expected: {}'.format(
                self.onu_device.mib_data_sync, value)
        return results

    def _check_stats(self, results, _, stat, expected):
        snapshot = self._snapshot_stats()
        assert snapshot[stat] == expected, \
            'Invalid statistic "{}". Got {}, Expected: {}'.format(stat,
                                                                  snapshot[stat],
                                                                  expected)
        return results

    def _check_value_equal(self, results, name, value, expected):
        assert value == expected, \
            'Value "{}" not equal. Got {}, Expected: {}'.format(name, value,
                                                                expected)
        return results

    def _default_errback(self, failure):
        from twisted.internet.defer import TimeoutError
        assert isinstance(failure.type, type(TimeoutError))

    def _snapshot_stats(self):
        omci_cc = self.onu_handler.omci_cc
        return {
            'tx_frames': omci_cc.tx_frames,
            'rx_frames': omci_cc.rx_frames,
            'rx_unknown_tid': omci_cc.rx_unknown_tid,
            'rx_onu_frames': omci_cc.rx_onu_frames,
            'rx_onu_discards': omci_cc.rx_onu_discards,
            'rx_timeouts': omci_cc.rx_timeouts,
            'rx_unknown_me': omci_cc.rx_unknown_me,
            'rx_late': omci_cc.rx_late,
            'tx_errors': omci_cc.tx_errors,
            'consecutive_errors': omci_cc.consecutive_errors,
            'reply_min': omci_cc.reply_min,
            'reply_max': omci_cc.reply_max,
            'reply_average': omci_cc.reply_average,
            'hp_tx_queue_len': omci_cc.hp_tx_queue_len,
            'lp_tx_queue_len': omci_cc.lp_tx_queue_len,
            'max_hp_tx_queue': omci_cc.max_hp_tx_queue,
            'max_lp_tx_queue': omci_cc._max_lp_tx_queue,
        }

    def test_default_init(self):
        self.setup_one_of_each()
        # Test default construction of OMCI_CC as well as
        # various other parameter settings
        omci_cc = self.onu_handler.omci_cc

        # No device directly associated
        self.assertIsNotNone(omci_cc._adapter_agent)
        self.assertIsNone(omci_cc._proxy_address)

        # No outstanding requests
        self.assertEqual(len(omci_cc._pending[OMCI_CC.LOW_PRIORITY]), 0)
        self.assertEqual(len(omci_cc._pending[OMCI_CC.HIGH_PRIORITY]), 0)

        # No active requests
        self.assertIsNone(omci_cc._tx_request[OMCI_CC.LOW_PRIORITY])
        self.assertIsNone(omci_cc._tx_request[OMCI_CC.HIGH_PRIORITY])

        # Flags/properties
        self.assertFalse(omci_cc.enabled)

        # Statistics
        self.assertEqual(omci_cc.tx_frames, 0)
        self.assertEqual(omci_cc.rx_frames, 0)
        self.assertEqual(omci_cc.rx_unknown_tid, 0)
        self.assertEqual(omci_cc.rx_onu_frames, 0)
        self.assertEqual(omci_cc.rx_onu_discards, 0)
        self.assertEqual(omci_cc.rx_unknown_me, 0)
        self.assertEqual(omci_cc.rx_timeouts, 0)
        self.assertEqual(omci_cc.rx_late, 0)
        self.assertEqual(omci_cc.tx_errors, 0)
        self.assertEqual(omci_cc.consecutive_errors, 0)
        self.assertNotEquals(omci_cc.reply_min, 0.0)
        self.assertEqual(omci_cc.reply_max, 0.0)
        self.assertEqual(omci_cc.reply_average, 0.0)
        self.assertEqual(omci_cc.lp_tx_queue_len, 0.0)
        self.assertEqual(omci_cc.max_hp_tx_queue, 0.0)
        self.assertEqual(omci_cc._max_hp_tx_queue, 0.0)
        self.assertEqual(omci_cc._max_lp_tx_queue, 0.0)

    def test_enable_disable(self):
        self.setup_one_of_each()

        # Test enable property
        omci_cc = self.onu_handler.omci_cc

        # Initially disabled
        self.assertFalse(omci_cc.enabled)
        omci_cc.enabled = False
        self.assertFalse(omci_cc.enabled)

        omci_cc.enabled = True
        self.assertTrue(omci_cc.enabled)
        self.assertIsNotNone(omci_cc._proxy_address)
        self.assertEqual(len(omci_cc._pending[OMCI_CC.LOW_PRIORITY]), 0)
        self.assertEqual(len(omci_cc._pending[OMCI_CC.HIGH_PRIORITY]), 0)

        omci_cc.enabled = True      # Should be a NOP
        self.assertTrue(omci_cc.enabled)
        self.assertIsNotNone(omci_cc._proxy_address)
        self.assertEqual(len(omci_cc._pending[OMCI_CC.LOW_PRIORITY]), 0)
        self.assertEqual(len(omci_cc._pending[OMCI_CC.HIGH_PRIORITY]), 0)

        omci_cc.enabled = False
        self.assertFalse(omci_cc.enabled)
        self.assertIsNone(omci_cc._proxy_address)

    def test_rx_discard_if_disabled(self):
        # ME without a known decoder
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = False
        snapshot = self._snapshot_stats()

        msg = '00fc2e0a00020000ff780000e00000010000000c' \
              '0000000000000000000000000000000000000000' \
              '00000028105a86ef'

        omci_cc.receive_message(hex2raw(msg))

        # Note: No counter increments
        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'])
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'])
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'])

    def test_message_send_get(self):
        # Various tests of sending an OMCI message and it either
        # getting a response or send catching some errors of
        # importance
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()
        mib_data_sync = self.onu_device.mib_data_sync

        # GET
        # d = omci_cc.send()  # TODO: Implement
        #
        # d.addCallbacks(self._is_omci_frame, self._default_errback)
        # d.addCallback(self._check_status, RC.Success.value)
        # d.addCallback(self._check_mib_sync, mib_data_sync)
        #
        # d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        # d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        # d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        # d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)

        # return d

    def test_message_send_set(self):
        # Various tests of sending an OMCI message and it either
        # getting a response or send catching some errors of
        # importance
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()
        mib_data_sync = self.onu_device.mib_data_sync

        # SET
        # d = omci_cc.send()  # TODO: Implement
        #
        # d.addCallbacks(self._is_omci_frame, self._default_errback)
        # d.addCallback(self._check_status, RC.Success.value)
        # d.addCallback(self._check_mib_sync, mib_data_sync + 1 if mib_data_sync < 255 else 1)
        #
        # d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        # d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        # d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        # d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)

        # return d
        #
        # # Also test mib_data_sync rollover.  255 -> 1  (zero reserved)
        #
        # self.onu_device.mib_data_sync = 255
        # # SET
        # self.assertTrue(True)  # TODO: Implement (copy previous one here)
        # self.assertEqual(1, self.onu_device.mib_data_sync)

    def test_message_send_create(self):
        # Various tests of sending an OMCI message and it either
        # getting a response or send catching some errors of
        # importance
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()
        mib_data_sync = self.onu_device.mib_data_sync

        # Create
        # d = omci_cc.send()  # TODO: Implement
        #
        # d.addCallbacks(self._is_omci_frame, self._default_errback)
        # d.addCallback(self._check_status, RC.Success.value)
        # d.addCallback(self._check_mib_sync, mib_data_sync + 1 if mib_data_sync < 255 else 1)
        #
        # d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        # d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        # d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        # d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)

        # return d

    def test_message_send_delete(self):
        # Various tests of sending an OMCI message and it either
        # getting a response or send catching some errors of
        # importance
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()
        mib_data_sync = self.onu_device.mib_data_sync

        # Delete
        # d = omci_cc.send()  # TODO: Implement
        #
        # d.addCallbacks(self._is_omci_frame, self._default_errback)
        # d.addCallback(self._check_status, RC.Success.value)
        # d.addCallback(self._check_mib_sync, mib_data_sync + 1 if mib_data_sync < 255 else 1)
        #
        # d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        # d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        # d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        # d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)

        # return d

    def test_message_send_mib_reset(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        self.onu_device.mib_data_sync = 10
        snapshot = self._snapshot_stats()

        # Successful MIB Reset
        d = omci_cc.send_mib_reset(timeout=1.0)

        d.addCallbacks(self._is_omci_frame, self._default_errback)
        d.addCallback(self._check_status, RC.Success)
        d.addCallback(self._check_mib_sync, 0)

        d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        d.addCallback(self._check_stats, snapshot, 'rx_late', snapshot['rx_late'])
        d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)
        return d

    def test_message_send_mib_upload(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()
        mib_data_sync = self.onu_device.mib_data_sync

        # MIB Upload
        d = omci_cc.send_mib_upload(timeout=1.0)

        d.addCallbacks(self._is_omci_frame, self._default_errback)
        d.addCallback(self._check_status, RC.Success)
        d.addCallback(self._check_mib_sync, mib_data_sync)

        # TODO: MIB Upload Results specific tests here

        d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        d.addCallback(self._check_stats, snapshot, 'rx_late', snapshot['rx_late'])
        d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)
        return d

    def test_message_send_mib_upload_next(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()
        mib_data_sync = self.onu_device.mib_data_sync

        # # MIB Upload Next
        # d = omci_cc.send_mib_upload_next(0, timeout=1.0)
        #
        # d.addCallbacks(self._is_omci_frame, self._default_errback)
        # d.addCallback(self._check_status, RC.Success)
        # d.addCallback(self._check_mib_sync, mib_data_sync)
        #
        # # TODO: MIB Upload Next Results specific tests here
        #
        # d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        # d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        # d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        # d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        # d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        # d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)
        # return d

    def test_message_send_no_timeout(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        self.onu_device.mib_data_sync = 10
        snapshot = self._snapshot_stats()

        d = omci_cc.send_mib_reset(timeout=0)
        d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'])
        d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        return d

    def test_message_send_bad_timeout(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        self.onu_device.mib_data_sync = 10
        snapshot = self._snapshot_stats()

        d = omci_cc.send_mib_reset(timeout=MAX_OMCI_REQUEST_AGE + 1)
        d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'])
        d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'])
        d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'] + 1)
        return d

    def test_message_send_not_a_frame(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        self.onu_device.mib_data_sync = 10
        snapshot = self._snapshot_stats()

        d = omci_cc.send('hello world', timeout=1)
        d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'])
        d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'])
        d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'] + 1)
        return d

    def test_message_send_reboot(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # ONU Reboot
        d = omci_cc.send_reboot(timeout=1.0)

        d.addCallbacks(self._is_omci_frame, self._default_errback)
        d.addCallback(self._check_status, RC.Success)

        d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 1)
        d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 1)
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'])
        d.addCallback(self._check_stats, snapshot, 'rx_late', snapshot['rx_late'])
        d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)
        return d

    def test_message_send_with_omci_disabled(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        self.assertFalse(omci_cc.enabled)

        # Successful MIB Reset
        d = omci_cc.send_mib_reset(timeout=1.0)

        def success_is_bad(_results):
            assert False, 'This test should throw a failure/error'

        def fail_fast(_failure):
            pass
            return None

        d.addCallbacks(success_is_bad, fail_fast)
        return d

    def test_message_send_get_with_latency(self):
        # Various tests of sending an OMCI message and it either
        # getting a response or send catching some errors of
        # importance
        self.setup_one_of_each()
        self.olt_handler.latency = 0.500    # 1/2 second

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True

        # Successful MIB Reset
        d = omci_cc.send_mib_reset(timeout=1.0)

        d.addCallbacks(self._is_omci_frame, self._default_errback)
        d.addCallback(self._check_status, RC.Success)

        def check_latency_values(_):
            self.assertGreaterEqual(omci_cc.reply_min, self.olt_handler.latency)
            self.assertGreaterEqual(omci_cc.reply_max, self.olt_handler.latency)
            self.assertGreaterEqual(omci_cc.reply_average, self.olt_handler.latency)

        d.addCallback(check_latency_values)
        return d

    def test_message_failures(self):
        # Various tests of sending an OMCI message and it fails
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        self.assertEqual(omci_cc.tx_frames, 0)
        self.assertEqual(omci_cc.rx_frames, 0)
        self.assertEqual(omci_cc.rx_unknown_tid, 0)
        self.assertEqual(omci_cc.rx_timeouts, 0)
        self.assertEqual(omci_cc.rx_late, 0)
        self.assertEqual(omci_cc.tx_errors, 0)

        # # Class ID not found
        # d = omci_cc.send_mib_reset(timeout=1.0)
        # self.assertTrue(True)  # TODO: Implement
        # todo: Test non-zero consecutive errors
        #
        # # Instance ID not found
        # d = omci_cc.send_mib_reset(timeout=1.0)
        # self.assertTrue(True)  # TODO: Implement
        # todo: Test non-zero consecutive errors
        #
        # # PON is disabled
        # d = omci_cc.send_mib_reset(timeout=1.0)
        # self.assertTrue(True)  # TODO: Implement
        # todo: Test non-zero consecutive errors
        #
        # # ONU is disabled
        # d = omci_cc.send_mib_reset(timeout=1.0)
        # self.assertTrue(True)  # TODO: Implement
        # todo: Test non-zero consecutive errors
        #
        # # ONU is not activated
        # d = omci_cc.send_mib_reset(timeout=1.0)
        # self.assertTrue(True)  # TODO: Implement
        # todo: Test non-zero consecutive errors

        # TODO: make OLT send back an unknown TID (

        # todo: Test non-zero consecutive errors
        # todo: Send a good frame
        # todo: Test zero consecutive errors
        # d.addCallback(self._check_value_equal, 'consecutive_errors', 0, omci_cc.consecutive_errors)

    def test_rx_unknown_me(self):
        # ME without a known decoder
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # This is the ID ------+
        #                      v
        msg = '00fc2e0a00020000ff780000e00000010000000c' \
              '0000000000000000000000000000000000000000' \
              '00000028'

        omci_cc.receive_message(hex2raw(msg))

        # Note: After successful frame decode, a lookup of the corresponding request by
        #       TID is performed. None should be found, so we should see the Rx Unknown TID
        #       increment.
        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'] + 1)
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'] + 1)
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'] + 1)
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'])
        self.assertEqual(omci_cc.consecutive_errors, 0)

    def test_rx_decode_unknown_me(self):
        # ME without a known decoder
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # This is a MIB Upload Next Response. Where we would probably first see an
        # unknown Class ID
        #
        # This is the ID ------+
        #                      v
        msg = '00fc2e0a00020000ff780001e000'
        blob = '00010000000c0000000000000000000000000000000000000000'
        msg += blob + '00000028'

        # Dig into the internal method so we can get the returned frame
        frame = omci_cc._decode_unknown_me(hex2raw(msg))

        self.assertEqual(frame.fields['transaction_id'], 0x00fc)
        self.assertEqual(frame.fields['message_type'], 0x2e)

        omci_fields = frame.fields['omci_message'].fields

        self.assertEqual(omci_fields['entity_class'], 0x0002)
        self.assertEqual(omci_fields['entity_id'], 0x00)
        self.assertEqual(omci_fields['object_entity_class'], 0x0ff78)
        self.assertEqual(omci_fields['object_entity_id'], 0x01)
        self.assertEqual(omci_fields['object_attributes_mask'], 0xe000)

        data_fields = omci_fields['object_data']

        decoded_blob = data_fields.get(UNKNOWN_CLASS_ATTRIBUTE_KEY)
        self.assertIsNotNone(decoded_blob)
        self.assertEqual(decoded_blob, blob)

    def test_rx_unknown_me_avc(self):
        # ME without a known decoder but is and attribute value change
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        msg = '0000110aff78000080000e000000' \
              '00000000000000000000000000000000000000000000000000000' \
              '00000028'

        omci_cc.receive_message(hex2raw(msg))

        # Blob decode should work and then it should be passed off to the
        # ONU Autonomous frame processor
        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'])
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'] + 1)
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'])
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'] + 1)
        self.assertEqual(omci_cc.rx_onu_discards, snapshot['rx_onu_discards'])
        self.assertEqual(omci_cc.consecutive_errors, 0)

    def test_rx_discard_if_disabled(self):
        # ME without a known decoder
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = False
        snapshot = self._snapshot_stats()

        msg = '00fc2e0a00020000ff780000e00000010000000c' \
              '0000000000000000000000000000000000000000' \
              '00000028105a86ef'

        omci_cc.receive_message(hex2raw(msg))

        # Note: No counter increments
        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'])
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'])
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'])

    def test_omci_alarm_decode(self):
        """
        This test covers an issue discovered in Sept 2018 (JIRA-1213).  It was
        an exception during frame decode.
        """
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # Frame from the JIRA issue
        msg = '0000100a000b0102800000000000000000000000' \
              '0000000000000000000000000000000000000015' \
              '000000282d3ae0a6'

        _results = omci_cc.receive_message(hex2raw(msg))

        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'])
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'])
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'] + 1)
        self.assertEqual(omci_cc.rx_onu_discards, snapshot['rx_onu_discards'])

    def test_omci_avc_decode(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # Frame from the JIRA issue
        msg = '0000110a0007000080004d4c2d33363236000000' \
              '0000000020202020202020202020202020202020' \
              '00000028'

        _results = omci_cc.receive_message(hex2raw(msg))

        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'])
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'])
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'] + 1)
        self.assertEqual(omci_cc.rx_onu_discards, snapshot['rx_onu_discards'])

    def test_omci_unknown_onu_decode(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # Frame from the JIRA issue
        msg = '0000190a0007000080004d4c2d33363236000000' \
              '0000000020202020202020202020202020202020' \
              '00000028'

        _results = omci_cc.receive_message(hex2raw(msg))

        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'])
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'])
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'] + 1)
        self.assertEqual(omci_cc.rx_onu_discards, snapshot['rx_onu_discards'] + 1)

    def test_omci_bad_frame_decode(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # Frame from the JIRA issue
        msg = '0020190a0007000080004d4c2d33363236000000' \
              '0000000000000028'

        _results = omci_cc.receive_message(hex2raw(msg))
        # NOTE: Currently do not increment any Rx Discard counters, just throw it away
        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'] + 1)
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'] + 1)
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'])
        self.assertEqual(omci_cc.rx_onu_discards, snapshot['rx_onu_discards'])

    def test_rx_decode_onu_g(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        msg = '001e2e0a0002000001000000e000424657530000' \
              '0000000000000000000000324246575300107496' \
              '00000028e7fb4a91'

        omci_cc.receive_message(hex2raw(msg))

        # Note: No counter increments
        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'] + 1)
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'] + 1)
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'])

    def test_rx_decode_extvlantagging(self):
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        msg = '030a290a00ab0201000d00000000001031323334' \
              '3536373839303132333435363738393031323334' \
              '000000281166d283'

        omci_cc.receive_message(hex2raw(msg))

        self.assertEqual(omci_cc.rx_frames, snapshot['rx_frames'] + 1)
        self.assertEqual(omci_cc.rx_unknown_me, snapshot['rx_unknown_me'])
        self.assertEqual(omci_cc.rx_unknown_tid, snapshot['rx_unknown_tid'] + 1)
        self.assertEqual(omci_cc.rx_onu_frames, snapshot['rx_onu_frames'])

    def _check_vlan_tag_op(self, results, attr, expected):
        omci_msg = results.fields['omci_message']
        data = omci_msg.fields['data']
        val = data[attr]
        self.assertEqual(expected, val)
        return results

    #@deferred()
    def test_rx_table_get_extvlantagging(self):
        self.setup_one_of_each()

        onu = self.onu_handler.onu_mock
        entity_id = 1
        vlan_tag_op1 = VlanTaggingOperation(
                                     filter_outer_priority=15,
                                     filter_outer_vid=4096,
                                     filter_outer_tpid_de=2,
                                     filter_inner_priority=15,
                                     filter_inner_vid=4096,
                                     filter_inner_tpid_de=0,
                                     filter_ether_type=0,
                                     treatment_tags_to_remove=0,
                                     treatment_outer_priority=15,
                                     treatment_outer_vid=1234,
                                     treatment_outer_tpid_de=0,
                                     treatment_inner_priority=0,
                                     treatment_inner_vid=4091,
                                     treatment_inner_tpid_de=4,
                                 )
        vlan_tag_op2 = VlanTaggingOperation(
                                     filter_outer_priority=14,
                                     filter_outer_vid=1234,
                                     filter_outer_tpid_de=5,
                                     filter_inner_priority=1,
                                     filter_inner_vid=2345,
                                     filter_inner_tpid_de=1,
                                     filter_ether_type=0,
                                     treatment_tags_to_remove=1,
                                     treatment_outer_priority=15,
                                     treatment_outer_vid=2222,
                                     treatment_outer_tpid_de=1,
                                     treatment_inner_priority=1,
                                     treatment_inner_vid=3333,
                                     treatment_inner_tpid_de=5,
                                 )
        vlan_tag_op3 = VlanTaggingOperation(
                                     filter_outer_priority=13,
                                     filter_outer_vid=55,
                                     filter_outer_tpid_de=1,
                                     filter_inner_priority=7,
                                     filter_inner_vid=4567,
                                     filter_inner_tpid_de=1,
                                     filter_ether_type=0,
                                     treatment_tags_to_remove=1,
                                     treatment_outer_priority=2,
                                     treatment_outer_vid=1111,
                                     treatment_outer_tpid_de=1,
                                     treatment_inner_priority=1,
                                     treatment_inner_vid=3131,
                                     treatment_inner_tpid_de=5,
                                 )
        tbl = [vlan_tag_op1, vlan_tag_op2, vlan_tag_op3]
        tblstr = str(vlan_tag_op1) + str(vlan_tag_op2) + str(vlan_tag_op3)

        onu._omci_response[OP.Get.value][ExtendedVlanTaggingOperationConfigurationData.class_id] = {
            entity_id: OmciFrame(transaction_id=0,
                         message_type=OmciGetResponse.message_id,
                         omci_message=OmciGetResponse(
                               entity_class=ExtendedVlanTaggingOperationConfigurationData.class_id,
                               entity_id=1,
                               success_code=RC.Success.value,
                               attributes_mask=ExtendedVlanTaggingOperationConfigurationData.mask_for(
                                   'received_frame_vlan_tagging_operation_table'),
                               data={'received_frame_vlan_tagging_operation_table': 16 * len(tbl)}
                         ))
        }

        rsp1 = binascii.a2b_hex(hexify(tblstr[0:OmciTableField.PDU_SIZE]))
        rsp2 = binascii.a2b_hex(hexify(tblstr[OmciTableField.PDU_SIZE:]))
        onu._omci_response[OP.GetNext.value][ExtendedVlanTaggingOperationConfigurationData.class_id] = {
            entity_id: {0: {'failures':2,
                            'frame':OmciFrame(transaction_id=0,
                                 message_type=OmciGetNextResponse.message_id,
                                 omci_message=OmciGetNextResponse(
                                     entity_class=ExtendedVlanTaggingOperationConfigurationData.class_id,
                                     entity_id=1,
                                     success_code=RC.Success.value,
                                     attributes_mask=ExtendedVlanTaggingOperationConfigurationData.mask_for(
                                         'received_frame_vlan_tagging_operation_table'),
                                     data={'received_frame_vlan_tagging_operation_table': rsp1
                                     }
                         ))},
                        1: OmciFrame(transaction_id=0,
                         message_type=OmciGetNextResponse.message_id,
                         omci_message=OmciGetNextResponse(
                             entity_class=ExtendedVlanTaggingOperationConfigurationData.class_id,
                             entity_id=1,
                             success_code=RC.Success.value,
                             attributes_mask=ExtendedVlanTaggingOperationConfigurationData.mask_for(
                                 'received_frame_vlan_tagging_operation_table'),
                             data={'received_frame_vlan_tagging_operation_table': rsp2
                             }
                         ))
                       }
        }

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True

        msg = ExtendedVlanTaggingOperationConfigurationDataFrame(
            entity_id,
            attributes={'received_frame_vlan_tagging_operation_table':True}
        )

        snapshot = self._snapshot_stats()

        frame = msg.get()
        d = omci_cc.send(frame, timeout=5.0)

        d.addCallbacks(self._is_omci_frame, self._default_errback, [OmciGetResponse])
        d.addCallback(self._check_status, RC.Success)

        d.addCallback(self._check_stats, snapshot, 'tx_frames', snapshot['tx_frames'] + 5)
        d.addCallback(self._check_stats, snapshot, 'rx_frames', snapshot['rx_frames'] + 3)
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_tid', snapshot['rx_unknown_tid'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_frames', snapshot['rx_onu_frames'])
        d.addCallback(self._check_stats, snapshot, 'rx_onu_discards', snapshot['rx_onu_discards'])
        d.addCallback(self._check_stats, snapshot, 'rx_unknown_me', snapshot['rx_unknown_me'])
        d.addCallback(self._check_stats, snapshot, 'rx_timeouts', snapshot['rx_timeouts'] + 2)
        d.addCallback(self._check_stats, snapshot, 'rx_late', snapshot['rx_late'])
        d.addCallback(self._check_stats, snapshot, 'tx_errors', snapshot['tx_errors'])
        d.addCallback(self._check_stats, snapshot, 'consecutive_errors', 0)
        d.addCallback(self._check_vlan_tag_op, 'received_frame_vlan_tagging_operation_table', tbl)

        return d

    ##################################################################
    # Start of tests specific to new stop_and_wait changes
    #
    def test_message_send_low_priority(self):
        # self.setup_one_of_each(timeout_messages=True)
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # MIB Upload
        d = omci_cc.send_mib_upload(timeout=1.0, high_priority=False)
        d.addCallback(self._check_stats, snapshot, 'lp_tx_queue_len', snapshot['lp_tx_queue_len'])
        d.addCallback(self._check_stats, snapshot, 'hp_tx_queue_len', snapshot['hp_tx_queue_len'])
        d.addCallback(self._check_stats, snapshot, 'max_lp_tx_queue', snapshot['max_lp_tx_queue'] + 1)
        d.addCallback(self._check_stats, snapshot, 'max_hp_tx_queue', snapshot['max_hp_tx_queue'])

        # Flush to get ready for next test (one frame queued)
        omci_cc.flush()
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        self.adapter_agent.timeout_the_message = True
        omci_cc.send_mib_upload(timeout=1.0, high_priority=False)
        omci_cc.send_mib_upload(timeout=1.0, high_priority=False)

        self.assertEqual(omci_cc.lp_tx_queue_len, 1)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)
        self.assertEqual(omci_cc.max_lp_tx_queue, 1)
        self.assertEqual(omci_cc.max_hp_tx_queue, 0)

        # Flush to get ready for next test (two queued and new max)
        omci_cc.flush()
        omci_cc.send_mib_upload(timeout=1.0, high_priority=False)
        omci_cc.send_mib_upload(timeout=1.0, high_priority=False)
        omci_cc.send_mib_upload(timeout=1.0, high_priority=False)

        self.assertEqual(omci_cc.lp_tx_queue_len, 2)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)
        self.assertEqual(omci_cc.max_lp_tx_queue, 2)
        self.assertEqual(omci_cc.max_hp_tx_queue, 0)

    def test_message_send_high_priority(self):
        # self.setup_one_of_each(timeout_messages=True)
        self.setup_one_of_each()

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True
        snapshot = self._snapshot_stats()

        # MIB Upload
        d = omci_cc.send_mib_upload(high_priority=True)
        d.addCallback(self._check_stats, snapshot, 'lp_tx_queue_len', snapshot['lp_tx_queue_len'])
        d.addCallback(self._check_stats, snapshot, 'hp_tx_queue_len', snapshot['hp_tx_queue_len'])
        d.addCallback(self._check_stats, snapshot, 'max_lp_tx_queue', snapshot['max_lp_tx_queue'])
        d.addCallback(self._check_stats, snapshot, 'max_hp_tx_queue', snapshot['max_hp_tx_queue'] + 1)

        # Flush to get ready for next test (one frame queued)
        omci_cc.flush()
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        self.adapter_agent.timeout_the_message = True
        omci_cc.send_mib_upload(high_priority=True)
        omci_cc.send_mib_upload(high_priority=True)

        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 1)
        self.assertEqual(omci_cc.max_lp_tx_queue, 0)
        self.assertEqual(omci_cc.max_hp_tx_queue, 1)

        # Flush to get ready for next test (two queued and new max)
        omci_cc.flush()
        omci_cc.send_mib_upload(high_priority=True)
        omci_cc.send_mib_upload(high_priority=True)
        omci_cc.send_mib_upload(high_priority=True)

        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 2)
        self.assertEqual(omci_cc.max_lp_tx_queue, 0)
        self.assertEqual(omci_cc.max_hp_tx_queue, 2)

    def test_message_send_and_cancel(self):
        global error_reason
        global successful
        # Do not send messages to adapter_agent
        self.setup_one_of_each(timeout_messages=True)

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True

        def success(_results):
            global successful
            successful = True

        def failure(reason):
            global error_reason
            error_reason = reason

        def notCalled(reason):
            assert isinstance(reason, Failure), 'Should not be called with success'

        # Cancel one that is actively being sent
        d = omci_cc.send_mib_upload(high_priority=False)
        d.addCallbacks(success, failure)
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        d.cancel()
        self.assertIsInstance(error_reason, Failure)
        self.assertFalse(successful)
        self.assertTrue(d.called)

        self.assertEqual(omci_cc.max_lp_tx_queue, 1)
        self.assertEqual(omci_cc.max_hp_tx_queue, 0)

        # Flush to get ready for next test (one running, one queued, cancel the
        # running one, so queued runs)
        omci_cc.flush()
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        error_reason = None
        d1 = omci_cc.send_mib_upload(high_priority=False)
        d2 = omci_cc.send_mib_upload(high_priority=False)
        d1.addCallbacks(success, failure)
        d2.addCallbacks(notCalled, notCalled)
        self.assertEqual(omci_cc.lp_tx_queue_len, 1)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        d1.cancel()
        self.assertIsInstance(error_reason, Failure)
        self.assertFalse(successful)
        self.assertTrue(d1.called)
        self.assertFalse(d2.called)

        self.assertEqual(omci_cc.max_lp_tx_queue, 1)
        self.assertEqual(omci_cc.max_hp_tx_queue, 0)
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        # Flush to get ready for next test (one running, one queued, cancel the queued one)

        omci_cc.flush()
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        error_reason = None
        d3 = omci_cc.send_mib_upload(timeout=55, high_priority=False)
        d4 = omci_cc.send_mib_upload(timeout=55, high_priority=False)
        d5 = omci_cc.send_mib_upload(timeout=55, high_priority=False)
        d3.addCallbacks(notCalled, notCalled)
        d4.addCallbacks(success, failure)
        d5.addCallbacks(notCalled, notCalled)
        self.assertEqual(omci_cc.lp_tx_queue_len, 2)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        d4.cancel()
        self.assertIsInstance(error_reason, Failure)
        self.assertFalse(successful)
        self.assertFalse(d3.called)
        self.assertTrue(d4.called)
        self.assertFalse(d5.called)

    def test_message_send_low_and_high_priority(self):
        self.setup_one_of_each(timeout_messages=True)

        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True

        omci_cc.send_mib_reset(high_priority=False)
        omci_cc.send_mib_reset(high_priority=True)
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        omci_cc.flush()
        self.assertEqual(omci_cc.lp_tx_queue_len, 0)
        self.assertEqual(omci_cc.hp_tx_queue_len, 0)

        omci_cc.send_mib_reset(high_priority=False)
        omci_cc.send_mib_reset(high_priority=True)
        omci_cc.send_mib_reset(high_priority=False)
        omci_cc.send_mib_reset(high_priority=True)
        self.assertEqual(omci_cc.lp_tx_queue_len, 1)
        self.assertEqual(omci_cc.hp_tx_queue_len, 1)

    def test_no_sw_download_and_mib_upload_at_same_time(self):
        # Section B.2.3 of ITU G.988-2017 specifies that a MIB
        # upload or software download at a given priority level
        # is not allowed while a similar action in the other
        # priority level is in progress. Relates to possible memory
        # consumption/needs on the ONU.
        #
        # OMCI_CC only checks if the commands are currently in
        # progress. ONU should reject messages if the upload/download
        # is in progress (but not an active request is in progress).

        self.setup_one_of_each(timeout_messages=True)
        omci_cc = self.onu_handler.omci_cc
        omci_cc.enabled = True

        mib_upload_msgs = [omci_cc.send_mib_upload,
                           # omci_cc.send_mib_upload_next
                           ]
        sw_download_msgs = [omci_cc.send_start_software_download,
                            # omci_cc.send_download_section,
                            # omci_cc.send_end_software_download
                            ]

        for upload in mib_upload_msgs:
            for download in sw_download_msgs:
                self.assertEqual(omci_cc.lp_tx_queue_len, 0)
                self.assertEqual(omci_cc.hp_tx_queue_len, 0)

                upload(high_priority=False)
                download(1, 1, 1, high_priority=True)    # Should stall send-next 50mS
                self.assertEqual(omci_cc.lp_tx_queue_len, 0)
                self.assertEqual(omci_cc.hp_tx_queue_len, 1)

                omci_cc.flush()
                self.assertEqual(omci_cc.lp_tx_queue_len, 0)
                self.assertEqual(omci_cc.hp_tx_queue_len, 0)

                upload(high_priority=True)
                download(1, 1, 1, high_priority=False)    # Should stall send-next 50mS
                self.assertEqual(omci_cc.lp_tx_queue_len, 1)
                self.assertEqual(omci_cc.hp_tx_queue_len, 0)

                omci_cc.flush()
                self.assertEqual(omci_cc.lp_tx_queue_len, 0)
                self.assertEqual(omci_cc.hp_tx_queue_len, 0)

                download(1, 1, 1, high_priority=False)
                upload(high_priority=True)    # Should stall send-next 50mS
                self.assertEqual(omci_cc.lp_tx_queue_len, 0)
                self.assertEqual(omci_cc.hp_tx_queue_len, 1)

                omci_cc.flush()
                self.assertEqual(omci_cc.lp_tx_queue_len, 0)
                self.assertEqual(omci_cc.hp_tx_queue_len, 0)

                download(1, 1, 1, high_priority=True)
                upload(high_priority=False)    # Should stall send-next 50mS)
                self.assertEqual(omci_cc.lp_tx_queue_len, 1)
                self.assertEqual(omci_cc.hp_tx_queue_len, 0)

                omci_cc.flush()
                self.assertEqual(omci_cc.lp_tx_queue_len, 0)
                self.assertEqual(omci_cc.hp_tx_queue_len, 0)

    # Some more ideas for tests that we could add
    # Send explicit tid that is not valid
    #       - Look at top of 'Send' method and test all the error conditions could may hit

    # Send multiple and have the OLT proxy throw an exception. Should call errback and
    # schedule remainder in queue to still tx.

    # Send a frame and then inject a response and test the RX logic out, including late
    # rx and retries by the OMCI_CC transmitter.


if __name__ == '__main__':
    main()



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import absolute_import
from .mock_adapter_agent import MockProxyAddress, MockDevice
from pyvoltha.adapters.extensions.omci.omci_cc import *
from pyvoltha.adapters.extensions.omci.omci_entities import entity_id_to_class_map


class MockOnuHandler(MockDevice):
    """
    Minimal class to handle ONU needs in OpenOMCI testing

    So that we do not have to duplicate the IAdapter functionality, just
    the handler, the OLT and ONU handlers are derived from a mock Device
    base class so that we can access the _devices map and get either a
    Device to play with (like the real thing) or the associated handler
    """
    def __init__(self, adapter_agent, parent_id, device_id, pon_id, onu_id):

        self.proxy_address = MockProxyAddress(parent_id, pon_id, onu_id)
        super(MockOnuHandler, self).__init__(device_id, self.proxy_address)

        self.device_id = device_id
        self.device = self
        self._adapter_agent = adapter_agent

        self.onu_mock = None
        self.omci_cc = OMCI_CC(adapter_agent, device_id, me_map=entity_id_to_class_map)

        # Items that you can change to perform various test failures

        self._enabled = True

    def tearDown(self):
        """Test case cleanup"""
        if self.onu_mock is not None:
            self.onu_mock.tearDown()
        self.onu_mock = None

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if self._enabled != value:
            self._enabled = value
            olt = self._adapter_agent.get_device(self.proxy_address.device_id)
            if olt is not None and self.proxy_address.channel_id in olt.enabled_pons:
                if self._enabled:
                    olt.activated_onus.add(self.serial_number)
                else:
                    olt.activated_onus.discard(self.serial_number)

    # Begin minimal set of needed IAdapter interfaces

    # TODO: Implement minimal functionality

    def send_proxied_message(self, proxy_address, msg):
        assert False, 'OpenOMCI will implement this for the MOCK ONU'

    def receive_proxied_message(self, _, msg):
        # Rx of OMCI message from MOCK OLT

        if self.omci_cc is not None and self.enabled:
            self.omci_cc.receive_message(msg.decode('hex'))



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Always prefer setuptools over distutils
from os import path
from setuptools import setup, find_packages

# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open

package = 'pyvoltha'
setup_dir = path.dirname(path.abspath(__file__))
version_file = path.join(setup_dir, "VERSION")

# Get the long description from the README file
with open(path.join(setup_dir, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open(version_file) as version_file:
    version = version_file.read().strip()

requirements = open(path.join(setup_dir, "requirements.txt")).read().splitlines()
required = [line for line in requirements if not line.startswith("-")]
print ("Required is '{}'".format(required))

setup(
    name=package,
    version=version,
    description='VOLTHA Python support libraries',
    author='Chip Boling',
    author_email='chip@bcsw.net',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://wiki.opencord.org/display/CORD/VOLTHA',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(exclude=['test']),
    install_requires=[required],
    include_package_data=True,
)



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from unittest import main, TestCase

from pyvoltha.adapters.extensions.omci.database.mib_db_ext import *
from pyvoltha.adapters.extensions.omci.database.mib_db_api import MODIFIED_KEY, CREATED_KEY,\
    DEVICE_ID_KEY, MDS_KEY, LAST_SYNC_KEY
from pyvoltha.adapters.extensions.omci.omci_cc import UNKNOWN_CLASS_ATTRIBUTE_KEY
from .mock.mock_adapter_agent import MockAdapterAgent, MockDevice
from nose.tools import raises, assert_raises
import time
import six

_DEVICE_ID = 'br-549'


class TestOmciMibDbExt(TestCase):

    def setUp(self):
        self.adapter_agent = MockAdapterAgent()
        self.adapter_agent.add_device(MockDevice(_DEVICE_ID))  # For Entity class lookups
        self.db = MibDbExternal(self.adapter_agent)

    def tearDown(self):
        self.db.stop()

    def test_start_stop(self):
        # Simple start stop
        self.assertFalse(self.db.active)
        self.db.start()
        self.assertTrue(self.db.active)
        self.db.stop()
        self.assertFalse(self.db.active)

        # Start after start still okay
        self.db.start()
        self.db.start()
        self.assertTrue(self.db.active)

        self.db.stop()
        self.db.stop()
        self.assertFalse(self.db.active)

    @raises(DatabaseStateError)
    def test_bad_state_add(self):
        self.db.add(_DEVICE_ID)

    @raises(DatabaseStateError)
    def test_bad_state_remove(self):
        self.db.remove(_DEVICE_ID)

    @raises(DatabaseStateError)
    def test_bad_state_query_1(self):
        self.db.query(_DEVICE_ID, 0)

    @raises(DatabaseStateError)
    def test_bad_state_query_2(self):
        self.db.query(_DEVICE_ID, 0, 0)

    @raises(DatabaseStateError)
    def test_bad_state_query_3(self):
        self.db.query(_DEVICE_ID, 0, 0, 'test')

    @raises(DatabaseStateError)
    def test_bad_state_set(self):
        self.db.set(_DEVICE_ID, 0, 0, {'test': 123})

    @raises(DatabaseStateError)
    def test_bad_state_delete(self):
        self.db.delete(_DEVICE_ID, 0, 0)

    @raises(KeyError)
    def test_no_device_query(self):
        self.db.start()
        self.db.query(_DEVICE_ID)

    def test_no_device_last_sync(self):
        self.db.start()
        # Returns None, not a KeyError
        value = self.db.get_last_sync(_DEVICE_ID)
        self.assertIsNone(value)

    def test_no_device_mds(self):
        self.db.start()
        # Returns None, not a KeyError
        value = self.db.get_mib_data_sync(_DEVICE_ID)
        self.assertIsNone(value)

    @raises(KeyError)
    def test_no_device_save_last_sync(self):
        self.db.start()
        self.db.save_last_sync(_DEVICE_ID, datetime.utcnow())

    @raises(KeyError)
    def test_no_device_save_mds(self):
        self.db.start()
        self.db.save_mib_data_sync(_DEVICE_ID, 123)

    def test_param_types(self):
        self.db.start()
        assert_raises(TypeError, self.db.add, 123)
        assert_raises(TypeError, self.db.remove, 123)
        assert_raises(TypeError, self.db.query, 123)

        assert_raises(TypeError, self.db.get_mib_data_sync, 123)
        assert_raises(TypeError, self.db.save_mib_data_sync, 123, 0)
        assert_raises(TypeError, self.db.save_mib_data_sync, _DEVICE_ID, 'zero')

        assert_raises(TypeError, self.db.get_last_sync, 123)
        assert_raises(TypeError, self.db.save_last_sync, 123, datetime.utcnow())
        assert_raises(TypeError, self.db.save_last_sync, _DEVICE_ID, 'bad-date')

        assert_raises(TypeError, self.db.set, 123, 0, 0, {'test': 0})
        assert_raises(TypeError, self.db.set, None, 0, 0, {'test': 0})
        assert_raises(ValueError, self.db.set, _DEVICE_ID, None, 0, {'test': 0})
        assert_raises(ValueError, self.db.set, _DEVICE_ID, 0, None, {'test': 0})
        assert_raises(TypeError, self.db.set, _DEVICE_ID, 0, 0, None)
        assert_raises(TypeError, self.db.set, _DEVICE_ID, 0, 0, 'not-a-dict')

        assert_raises(ValueError, self.db.set, _DEVICE_ID, -1, 0, {'test': 0})
        assert_raises(ValueError, self.db.set, _DEVICE_ID, 0x10000, 0, {'test': 0})
        assert_raises(ValueError, self.db.set, _DEVICE_ID, 0, -1, {'test': 0})
        assert_raises(ValueError, self.db.set, _DEVICE_ID, 0, 0x10000, {'test': 0})

        assert_raises(TypeError, self.db.delete, 123, 0, 0)
        assert_raises(ValueError, self.db.delete, _DEVICE_ID, -1, 0)
        assert_raises(ValueError, self.db.delete, _DEVICE_ID, 0x10000, 0)
        assert_raises(ValueError, self.db.delete, _DEVICE_ID, 0, -1)
        assert_raises(ValueError, self.db.delete, _DEVICE_ID, 0, 0x10000)

    def test_add_remove_device(self):
        self.db.start()

        # Remove of non-existent device is not an error
        assert_raises(KeyError, self.db.query, _DEVICE_ID)
        self.db.remove(_DEVICE_ID)

        start_time = datetime.utcnow()
        self.db.add(_DEVICE_ID)
        dev_data = self.db.query(_DEVICE_ID)

        self.assertEqual(dev_data[DEVICE_ID_KEY], _DEVICE_ID)
        self.assertEquals(dev_data[MDS_KEY], 0)
        self.assertIsNone(dev_data[LAST_SYNC_KEY])
        self.assertEqual(dev_data[VERSION_KEY], MibDbExternal.CURRENT_VERSION)

        self.assertGreaterEqual(self.db.created, start_time)

        # Remove it
        self.db.remove(_DEVICE_ID)
        assert_raises(KeyError, self.db.query, _DEVICE_ID)

        # Remove of non-existant dev okay
        self.db.remove(_DEVICE_ID +'abcd')

        # Overwrite tests
        self.db.add(_DEVICE_ID)
        assert_raises(KeyError, self.db.add, _DEVICE_ID)
        self.db.add(_DEVICE_ID, overwrite=True)  # This is okay

    def test_mib_data_sync(self):
        self.db.start()
        self.db.add(_DEVICE_ID)
        self.assertEquals(self.db.get_mib_data_sync(_DEVICE_ID), 0)

        self.db.save_mib_data_sync(_DEVICE_ID, 100)
        self.assertEqual(self.db.get_mib_data_sync(_DEVICE_ID), 100)

        assert_raises(ValueError, self.db.save_mib_data_sync, _DEVICE_ID, -1)
        assert_raises(ValueError, self.db.save_mib_data_sync, _DEVICE_ID, 256)

    def test_last_sync(self):
        self.db.start()
        self.assertIsNone(self.db.get_last_sync(_DEVICE_ID))

        self.db.add(_DEVICE_ID)
        self.assertIsNone(self.db.get_last_sync(_DEVICE_ID))

        now = datetime.utcnow()

        self.db.save_last_sync(_DEVICE_ID, now)
        self.assertEqual(self.db.get_last_sync(_DEVICE_ID), now)

        assert_raises(TypeError, self.db.save_last_sync, _DEVICE_ID, 'hello')

    def test_set_and_query(self):
        self.db.start()
        self.db.add(_DEVICE_ID)     # Base device DB created here
        time.sleep(0.1)

        class_id = OntG.class_id
        inst_id = 0
        attributes = {'vendor_id': 'ABCD'}

        start_time = datetime.utcnow()
        set_occurred = self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        self.assertTrue(set_occurred)
        end_time = datetime.utcnow()

        dev_data = self.db.query(_DEVICE_ID)
        self.assertEqual(dev_data[DEVICE_ID_KEY], _DEVICE_ID)

        dev_classes = [v for k, v in dev_data.items() if isinstance(k, int)]

        self.assertEqual(len(dev_classes), 1)
        class_data = dev_classes[0]

        self.assertEqual(class_data[CLASS_ID_KEY], class_id)

        class_insts = [v for k, v in class_data.items() if isinstance(k, int)]

        self.assertEqual(len(class_insts), 1)
        inst_data = class_insts[0]

        self.assertEqual(inst_data[INSTANCE_ID_KEY], inst_id)
        self.assertGreaterEqual(inst_data[MODIFIED_KEY], start_time)
        self.assertLessEqual(inst_data[MODIFIED_KEY], end_time)
        self.assertLessEqual(inst_data[CREATED_KEY], inst_data[MODIFIED_KEY])

        inst_attributes = inst_data[ATTRIBUTES_KEY]
        self.assertEqual(len(inst_attributes), 1)

        self.assertTrue('vendor_id' in inst_attributes)
        self.assertEqual(inst_attributes['vendor_id'], attributes['vendor_id'])

        ########################################
        # Query with device and class. Should be same as from full device query
        cls_2_data = self.db.query(_DEVICE_ID, class_id)

        self.assertEqual(class_data[CLASS_ID_KEY], cls_2_data[CLASS_ID_KEY])

        cl2_insts = {k:v for k, v in cls_2_data.items() if isinstance(k, int)}
        self.assertEqual(len(cl2_insts), len(class_insts))

        # Bad class id query
        cls_no_data = self.db.query(_DEVICE_ID, class_id + 1)
        self.assertTrue(isinstance(cls_no_data, dict))
        self.assertEqual(len(cls_no_data), 0)

        ########################################
        # Query with device, class, instance
        inst_2_data = self.db.query(_DEVICE_ID, class_id, inst_id)

        self.assertEqual(inst_data[INSTANCE_ID_KEY], inst_2_data[INSTANCE_ID_KEY])
        self.assertEqual(inst_data[MODIFIED_KEY], inst_2_data[MODIFIED_KEY])
        self.assertEqual(inst_data[CREATED_KEY], inst_2_data[CREATED_KEY])

        inst2_attr = inst_2_data[ATTRIBUTES_KEY]
        self.assertEqual(len(inst2_attr), len(inst_attributes))

        # Bad instance id query
        inst_no_data = self.db.query(_DEVICE_ID, class_id, inst_id + 100)
        self.assertTrue(isinstance(inst_no_data, dict))
        self.assertEqual(len(inst_no_data), 0)

        ########################################
        # Attribute queries
        attr_2_data = self.db.query(_DEVICE_ID, class_id, inst_id, 'vendor_id')
        self.assertEqual(attr_2_data['vendor_id'], attributes['vendor_id'])

        attr_3_data = self.db.query(_DEVICE_ID, class_id, inst_id, ['vendor_id'])
        self.assertEqual(attr_3_data['vendor_id'], attributes['vendor_id'])

        attr_4_data = self.db.query(_DEVICE_ID, class_id, inst_id, {'vendor_id'})
        self.assertEqual(attr_4_data['vendor_id'], attributes['vendor_id'])

        attr_no_data = self.db.query(_DEVICE_ID, class_id, inst_id, 'no_such_thing')
        self.assertTrue(isinstance(attr_no_data, dict))
        self.assertEqual(len(attr_no_data), 0)

        # Set to same value does not change modified data.  The modified is
        # at the instance level

        class_id = OntG.class_id
        inst_id = 0
        attributes = {'vendor_id': 'ABCD'}
        set_occurred = self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        self.assertFalse(set_occurred)

        inst_3_data = self.db.query(_DEVICE_ID, class_id, inst_id)
        self.assertEqual(inst_data[MODIFIED_KEY], inst_3_data[MODIFIED_KEY])
        self.assertEqual(inst_data[CREATED_KEY], inst_3_data[CREATED_KEY])

        # But set to new value does
        time.sleep(0.1)
        attributes = {'vendor_id': 'WXYZ'}
        set_occurred = self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        self.assertTrue(set_occurred)

        inst_4_data = self.db.query(_DEVICE_ID, class_id, inst_id)
        self.assertLess(inst_3_data[MODIFIED_KEY], inst_4_data[MODIFIED_KEY])
        self.assertEqual(inst_3_data[CREATED_KEY], inst_4_data[CREATED_KEY])

    def test_delete_instances(self):
        self.db.start()
        self.db.add(_DEVICE_ID)
        create_time = datetime.utcnow()

        class_id = GalEthernetProfile.class_id
        inst_id_1 = 0x100
        inst_id_2 = 0x200
        attributes = {'max_gem_payload_size': 1500}

        self.db.set(_DEVICE_ID, class_id, inst_id_1, attributes)
        self.db.set(_DEVICE_ID, class_id, inst_id_2, attributes)
        time.sleep(0.1)

        dev_data = self.db.query(_DEVICE_ID)
        cls_data = self.db.query(_DEVICE_ID, class_id)
        inst_data = {k: v for k, v in cls_data.items() if isinstance(k, int)}
        self.assertEqual(len(inst_data), 2)

        self.assertLessEqual(dev_data[CREATED_KEY], create_time)
        self.assertLessEqual(self.db.created, create_time)

        # Delete one instance
        time.sleep(0.1)
        result = self.db.delete(_DEVICE_ID, class_id, inst_id_1)
        self.assertTrue(result)     # True returned if a del actually happened

        dev_data = self.db.query(_DEVICE_ID)
        cls_data = self.db.query(_DEVICE_ID, class_id)
        inst_data = {k: v for k, v in cls_data.items() if isinstance(k, int)}
        self.assertEqual(len(inst_data), 1)

        self.assertLessEqual(dev_data[CREATED_KEY], create_time)
        self.assertLessEqual(self.db.created, create_time)

        # Delete remaining instance
        time.sleep(0.1)
        result = self.db.delete(_DEVICE_ID, class_id, inst_id_2)
        self.assertTrue(result)     # True returned if a del actually happened

        dev_data = self.db.query(_DEVICE_ID)
        cls_data = {k: v for k, v in dev_data.items() if isinstance(k, int)}
        self.assertEqual(len(cls_data), 0)
        self.assertLessEqual(dev_data[CREATED_KEY], create_time)

        # Delete returns false if not instance
        self.assertFalse(self.db.delete(_DEVICE_ID, class_id, inst_id_1))
        self.assertFalse(self.db.delete(_DEVICE_ID, class_id, inst_id_2))

    def test_on_mib_reset_listener(self):
        self.db.start()
        self.db.add(_DEVICE_ID)
        time.sleep(0.1)

        class_id = OntG.class_id
        inst_id = 0
        attributes = {'vendor_id': 'ABCD'}

        set_time = datetime.utcnow()
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)

        time.sleep(0.1)
        self.db.on_mib_reset(_DEVICE_ID)

        dev_data = self.db.query(_DEVICE_ID)
        self.assertEqual(dev_data[DEVICE_ID_KEY], _DEVICE_ID)
        self.assertLessEqual(dev_data[CREATED_KEY], set_time)
        self.assertLessEqual(self.db.created, set_time)

        self.assertFalse(any(isinstance(cls, int) for cls in six.iterkeys(dev_data)))

    def test_str_field_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = OltG.class_id
        inst_id = 0
        attributes = {
            'olt_vendor_id': 'ABCD',             # StrFixedLenField(4)
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(all(isinstance(data[k], six.string_types) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))

    def test_mac_address_ip_field_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = IpHostConfigData.class_id
        inst_id = 0
        attributes = {
            'mac_address': '00:01:02:03:04:05',             # MACField
            'ip_address': '1.2.3.4',                        # IPField
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(all(isinstance(data[k], six.string_types) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))

    def test_byte_and_short_field_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = UniG.class_id
        inst_id = 0
        attributes = {
            'administrative_state': int(1),                # ByteField
            'non_omci_management_identifier': int(12345)   # IPField
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(all(isinstance(data[k], type(attributes[k])) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))

    def test_int_field_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = PriorityQueueG.class_id
        inst_id = 0
        attributes = {
            'related_port': int(1234567)    # IntField
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(all(isinstance(data[k], type(attributes[k])) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))

    def test_long_field_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = PriorityQueueG.class_id
        inst_id = 0
        attributes = {
            'packet_drop_queue_thresholds': int(0x1234)        # LongField
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(all(isinstance(data[k], type(attributes[k])) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))

    def test_bit_field_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = OntG.class_id
        inst_id = 0
        attributes = {
            'extended_tc_layer_options': int(0x1234),        # BitField(16)
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(all(isinstance(data[k], type(attributes[k])) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))

    def test_list_field_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = VlanTaggingFilterData.class_id
        inst_id = 0
        vlan_filter_list = [0] * 12
        vlan_filter_list[0] = 0x1234

        attributes = {
            'vlan_filter_list': vlan_filter_list,        # FieldListField
            'forward_operation': 0,
            'number_of_entries': 1
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)
        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(all(isinstance(data[k], type(attributes[k])) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))

    def test_complex_json_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        class_id = ExtendedVlanTaggingOperationConfigurationData.class_id
        inst_id = 0x202
        table_data = VlanTaggingOperation(
            filter_outer_priority=15,
            filter_inner_priority=8,
            filter_inner_vid=1024,
            filter_inner_tpid_de=5,
            filter_ether_type=0,
            treatment_tags_to_remove=1,
            pad3=2,
            treatment_outer_priority=15,
            treatment_inner_priority=8,
            treatment_inner_vid=1024,
            treatment_inner_tpid_de=4
        )
        attributes = dict(
            received_frame_vlan_tagging_operation_table=table_data
        )
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)

        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        table_as_dict = json.loads(table_data.to_json())

        self.assertTrue(all(isinstance(data['received_frame_vlan_tagging_operation_table'][0].fields[k],
                                       type(attributes['received_frame_vlan_tagging_operation_table'].fields[k]))
                            for k in attributes['received_frame_vlan_tagging_operation_table'].fields.keys()))
        self.assertTrue(all(data['received_frame_vlan_tagging_operation_table'][0].fields[k] ==
                            attributes['received_frame_vlan_tagging_operation_table'].fields[k]
                            for k in attributes['received_frame_vlan_tagging_operation_table'].fields.keys()))
        self.assertTrue(all(data['received_frame_vlan_tagging_operation_table'][0].fields[k] == table_as_dict[k]
                            for k in table_as_dict.keys()))

    def test_unknown_me_serialization(self):
        self.db.start()
        self.db.add(_DEVICE_ID)

        blob = '00010000000c0000000000000000000000000000000000000000'
        class_id = 0xff78
        inst_id = 0x101
        attributes = {
            UNKNOWN_CLASS_ATTRIBUTE_KEY: blob
        }
        self.db.set(_DEVICE_ID, class_id, inst_id, attributes)

        data = self.db.query(_DEVICE_ID, class_id, inst_id, list(attributes.keys()))
        self.assertTrue(isinstance(UNKNOWN_CLASS_ATTRIBUTE_KEY, six.string_types))
        self.assertTrue(all(isinstance(attributes[k], six.string_types) for k in attributes.keys()))
        self.assertTrue(all(data[k] == attributes[k] for k in attributes.keys()))


if __name__ == '__main__':
    main()



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import absolute_import
import sys
from .mock_adapter_agent import MockDevice
from nose.twistedtools import reactor
from six.moves import range


class MockOltHandler(MockDevice):
    """
    VERY Minimal class to handle OLT needs in OpenOMCI testing

    So that we do not have to duplicate the IAdapter functionality, just
    the handler, the OLT and ONU handlers are derived from a mock Device
    base class so that we can access the _devices map and get either a
    Device to play with (like the real thing) or the associated handler
    """
    def __init__(self, adapter_agent, device_id):
        super(MockOltHandler, self).__init__(device_id)

        self.device_id = device_id
        self.device = self
        self._adapter_agent = adapter_agent
        self._num_tx = 0

        ####################################################################
        # NOTE: The following can be manipulated in your test case to modify the behaviour
        #       of this mock.
        #
        # Note that activated ONUs are added during adapter add_child_device
        # if the ONU handler associated is 'enabled'

        self.enabled = True                # OLT is enabled/active
        self.activated_onus = set()        # Activated ONU serial numbers
        self.enabled_pons = list(range(0, 16))   # Enabled PONs
        self.max_tx = sys.maxint           # Fail after this many tx requests
        self.latency = 0.0                 # OMCI response latency (keep small)

    # TODO: Implement minimal functionality

    # TODO: Implement minimal functionality

    def tearDown(self):
        """Test case cleanup"""
        pass

    # Begin minimal set of needed IAdapter interfaces

    def send_proxied_message(self, proxy_address, msg):
        """Check various enabled flags and status and send if okay"""

        if not self.enabled:
            return None

        pon_id = proxy_address.channel_id

        if pon_id not in self.enabled_pons:
            return None

        # Look up ONU device ID.
        onu_id = proxy_address.onu_id
        onu_handler = self._adapter_agent.get_child_device(proxy_address.device_id,
                                                           pon_id=pon_id,
                                                           onu_id=onu_id)

        if onu_handler is None or not onu_handler.enabled:
            return None

        onu_mock = onu_handler.onu_mock
        if onu_mock is None or onu_mock.serial_number not in self.activated_onus:
            return None

        # And Tx success (silent discard for OMCI timeout testing)
        if self._num_tx >= self.max_tx:
            return None
        self._num_tx += 1

        response = onu_mock.rx_omci_frame(msg)

        # Make async and add any requested latency. Bound it to less
        # than 5 seconds since this is a unit test that need to be
        # somewhat responsive

        assert 0.0 <= self.latency <= 5, 'Best practice is latency <= 5 seconds'
        if response is not None:
            reactor.callLater(self.latency, self._deliver_proxy_message, proxy_address, response)

    def _deliver_proxy_message(self, proxy_address, response):
        from common.frameio.frameio import hexify
        self._adapter_agent.receive_proxied_message(proxy_address,
                                                    hexify(response))

    def receive_proxied_message(self, _, __):
        assert False, 'This is never called on the OLT side of proxy messaging'




#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
import structlog
from enum import Enum

from pyvoltha.common.config.config_txn import ConfigTransaction

log = structlog.get_logger()


class OperationContext(object):
    def __init__(self, path=None, data=None, field_name=None, child_key=None):
        self.path = path
        self._data = data
        self.field_name = field_name
        self.child_key = child_key
    @property
    def data(self):
        return self._data
    def update(self, data):
        self._data = data
        return self
    def __repr__(self):
        return 'OperationContext({})'.format(self.__dict__)


class CallbackType(Enum):

    # GET hooks are called after the data is retrieved and can be used to
    # augment the data (they should only augment fields marked as REAL_TIME
    GET = 1

    # PRE_UPDATE hooks are called before the change is made and are supposed
    # to be used to reject the data by raising an exception. If they don't,
    # the change will be applied.
    PRE_UPDATE = 2

    # POST_UPDATE hooks are called after the update has occurred and can
    # be used to deal with the change. For instance, an adapter can use the
    # callback to trigger the south-bound configuration
    POST_UPDATE = 3

    # These behave similarly to the update callbacks as described above.
    PRE_ADD = 4
    POST_ADD = 5
    PRE_REMOVE = 6
    POST_REMOVE = 7

    # Bulk list change due to transaction commit that changed items in
    # non-keyed container fields
    POST_LISTCHANGE = 8


class ConfigProxy(object):
    """
    Allows an entity to look at a sub-tree and see it as it was the whole tree
    """
    __slots__ = (
        '_root',
        '_node',
        '_path',
        '_exclusive',
        '_callbacks'
    )

    def __init__(self, root, node, path, exclusive):
        self._root = root
        self._node = node
        self._exclusive = exclusive
        self._path = path  # full path to proxied node
        self._callbacks = {}  # call back type -> list of callbacks

    @property
    def exclusive(self):
        return self._exclusive

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~ CRUD handlers ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def get(self, path='/', depth=None, deep=None, txid=None):
        return self._node.get(path, depth=depth, deep=deep, txid=txid)

    def update(self, path, data, strict=False, txid=None):
        assert path.startswith('/')
        full_path = self._path if path == '/' else self._path + path
        return self._root.update(full_path, data, strict, txid=txid)

    def add(self, path, data, txid=None):
        assert path.startswith('/')
        full_path = self._path if path == '/' else self._path + path
        return self._root.add(full_path, data, txid=txid)

    def remove(self, path, txid=None):
        assert path.startswith('/')
        full_path = self._path if path == '/' else self._path + path
        return self._root.remove(full_path, txid=txid)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~ Transaction support ~~~~~~~~~~~~~~~~~~~~~~~~~~

    def open_transaction(self):
        """Open a new transaction"""
        txid = self._root.mk_txbranch()
        return ConfigTransaction(self, txid)

    def commit_transaction(self, txid):
        """
        If having an open transaction, commit it now. Will raise exception
        if conflict is detected. Either way, transaction will be deleted.
        """
        self._root.fold_txbranch(txid)

    def cancel_transaction(self, txid):
        """
        Cancel current transaction if we are in a transaction. Always succeeds.
        """
        self._root.del_txbranch(txid)

    # ~~~~~~~~~~~~~~~~~~~~~~ Callbacks registrations ~~~~~~~~~~~~~~~~~~~~~~~~~~

    def register_callback(self, callback_type, callback, *args, **kw):
        lst = self._callbacks.setdefault(callback_type, [])
        lst.append((callback, args, kw))

    def unregister_callback(self, callback_type, callback, *args, **kw):
        lst = self._callbacks.setdefault(callback_type, [])
        if (callback, args, kw) in lst:
            lst.remove((callback, args, kw))

    # ~~~~~~~~~~~~~~~~~~~~~ Callback dispatch ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def invoke_callbacks(self, callback_type, context, proceed_on_errors=False):
        lst = self._callbacks.get(callback_type, [])
        for callback, args, kw in lst:
            try:
                context = callback(context, *args, **kw)
            except Exception as e:
                if proceed_on_errors:
                    log.exception(
                        'call-back-error', callback_type=callback_type,
                        context=context, e=e)
                else:
                    raise
        return context



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
import structlog
from .task import Task
from twisted.internet import reactor
from voltha_protos.voltha_pb2 import ImageDownload


class OmciSwImageUpgradeTask(Task):
    name = "OMCI Software Image Upgrade Task"

    def __init__(self, img_id, omci_upgrade_sm_cls, omci_agent, image_download, clock=None):
        super(OmciSwImageUpgradeTask, self).__init__(OmciSwImageUpgradeTask.name, omci_agent, image_download.id,
                                                     exclusive=False,
                                                     watchdog_timeout=45)
        self.log.debug("OmciSwImageUpgradeTask create ", image_id=img_id)
        self._image_id = img_id
        self._omci_upgrade_sm_cls = omci_upgrade_sm_cls
        # self._omci_agent = omci_agent
        self._image_download = image_download
        self.reactor = clock if clock is not None else reactor
        self._omci_upgrade_sm = None
        self.log.debug("OmciSwImageUpgradeTask create end", image_id=img_id)

    @property 
    def status(self):
        return self._image_download
        
    def start(self):
        self.log.debug("OmciSwImageUpgradeTask start")
        super(OmciSwImageUpgradeTask, self).start()
        if self._omci_upgrade_sm is None:
            self._omci_upgrade_sm = self._omci_upgrade_sm_cls(self._image_id, self.omci_agent, self._image_download, clock=self.reactor)
            d = self._omci_upgrade_sm.start()
            d.chainDeferred(self.deferred)
        #else:
        #    if restart:
        #        self._omci_upgrade_sm.reset_image()

    def stop(self):
        self.log.debug("OmciSwImageUpgradeTask stop")
        if self._omci_upgrade_sm is not None:
            self._omci_upgrade_sm.stop()
            self._omci_upgrade_sm = None
    
    def onu_bootup(self):
        self.log.debug("onu_bootup", state=self._omci_upgrade_sm.status.image_state);
        if self._omci_upgrade_sm is not None \
            and self._omci_upgrade_sm.status.image_state == ImageDownload.IMAGE_ACTIVATE:
            self._omci_upgrade_sm.do_commit()
    



#
# Copyright 2020 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from .mib_db_api import CREATED_KEY, MODIFIED_KEY
import json
from datetime import datetime
import structlog
from pyvoltha.common.utils.registry import registry
from pyvoltha.common.config.config_backend import EtcdStore
import six


class MibTemplateDb(object):

    BASE_PATH = 'service/voltha/omci_mibs/templates'
    TEMPLATE_PATH = '{}/{}/{}'

    def __init__(self, vendor_id, equipment_id, software_version, serial_number, mac_address):
        self.log = structlog.get_logger()
        self._jsonstring = b''

        # lookup keys
        self._vendor_id = vendor_id
        self._equipment_id = equipment_id
        self._software_version = software_version

        # replacement values
        self._serial_number = serial_number
        self._mac_address = mac_address

        self.args = registry('main').get_args()
        host, port = self.args.etcd.split(':', 1)
        self._kv_store = EtcdStore(host, port, MibTemplateDb.BASE_PATH)
        self.loaded = False
        self._load_template()

    def get_template_instance(self):
        # swap out tokens with specific data
        fixup = self._jsonstring.decode('ascii')
        fixup = fixup.replace('%SERIAL_NUMBER%', self._serial_number)
        fixup = fixup.replace('%MAC_ADDRESS%', self._mac_address)

        # convert to a dict() compatible with mib_db_dict
        newdb = self._load_from_json(fixup)
        now = datetime.utcnow()

        # populate timestamps as if it was mib uploaded
        for cls_id, cls_data in newdb.items():
            if isinstance(cls_id, int):
                for inst_id, inst_data in cls_data.items():
                    if isinstance(inst_id, int):
                        newdb[cls_id][inst_id][CREATED_KEY] = now
                        newdb[cls_id][inst_id][MODIFIED_KEY] = now

        return newdb

    def _load_template(self):
        path = self._get_template_path()
        try:
            self._jsonstring = self._kv_store[path]
            self.log.debug('found-template-data', path=path)
            self.loaded = True
        except KeyError:
            self.log.warn('no-template-found', path=path)

    def _get_template_path(self):
        if not isinstance(self._vendor_id, six.string_types):
            raise TypeError('Vendor ID is a string')

        if not isinstance(self._equipment_id, six.string_types):
            raise TypeError('Equipment ID is a string')

        if not isinstance(self._software_version, six.string_types):
            raise TypeError('Software Version is a string')

        fmt = MibTemplateDb.TEMPLATE_PATH
        return fmt.format(self._vendor_id, self._equipment_id, self._software_version)

    def _load_from_json(self, jsondata):

        def json_obj_parser(x):
            if isinstance(x, dict):
                results = dict()
                for (key, value) in x.items():
                    try:
                        key = int(key)
                    except (ValueError, TypeError):
                        pass

                    results.update({key: value})
                return results
            return x

        template_data = json.loads(jsondata, object_hook=json_obj_parser)
        return template_data



# Copyright 2017-present Open Networking Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from unittest import TestCase, main
from pyvoltha.common.utils.indexpool import IndexPool
from six.moves import range

class TestIndexPool(TestCase):
    pool = IndexPool(8, 0)
    def test_01_get_next(self):
        self.assertEqual(self.pool.indices.bin, '00000000')
        for i in range(8):
            self.assertEqual(self.pool.get_next(), i)
        #to check if there's any bit left after using all 8 bits
        self.assertIsNone(self.pool.get_next())

    def test_02_pre_allocate(self):
        _pool2 = IndexPool(8, 0)
        self.assertEqual(_pool2.indices.bin, '00000000')
        _pool2.pre_allocate((0,1,2,))
        self.assertEqual(_pool2.indices.bin, '11100000')

    def test_03_release(self):
        self.pool.release(5)
        self.assertEqual(self.pool.indices.bin, '11111011')
        self.pool.release(10)
        self.assertEqual(self.pool.indices.bin, '11111011')
        self.pool.release(0)
        self.assertEqual(self.pool.indices.bin, '01111011')

    def test_04_check_offset(self):
        _offset = 5
        self.pool = IndexPool(8, _offset)
        for i in range(8):
            self.assertEqual(self.pool.get_next(), _offset + i)


if __name__ == '__main__':
    main()


#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Some network related convenience functions
"""

from __future__ import absolute_import
from __future__ import print_function
from netifaces import AF_INET

import netifaces as ni
import netaddr


def _get_all_interfaces():
    m_interfaces = []
    for iface in ni.interfaces():
        m_interfaces.append((iface, ni.ifaddresses(iface)))
    return m_interfaces


def _get_my_primary_interface():
    gateways = ni.gateways()
    assert 'default' in gateways, \
        ("No default gateway on host/container, "
         "cannot determine primary interface")
    default_gw_index = list(gateways['default'].keys())[0]
    # gateways[default_gw_index] has the format (example):
    # [('10.15.32.1', 'en0', True)]
    interface_name = gateways[default_gw_index][0][1]
    return interface_name


def get_my_primary_local_ipv4(inter_core_subnet=None, ifname=None):
    if not inter_core_subnet:
        return _get_my_primary_local_ipv4(ifname)
    # My IP should belong to the specified subnet
    for iface in ni.interfaces():
        addresses = ni.ifaddresses(iface)
        if AF_INET in addresses:
            m_ip = addresses[AF_INET][0]['addr']
            _ip = netaddr.IPAddress(m_ip).value
            m_network = netaddr.IPNetwork(inter_core_subnet)
            if _ip >= m_network.first and _ip <= m_network.last:
                return m_ip
    return None


def get_my_primary_interface(pon_subnet=None):
    if not pon_subnet:
        return _get_my_primary_interface()
    # My interface should have an IP that belongs to the specified subnet
    for iface in ni.interfaces():
        addresses = ni.ifaddresses(iface)
        if AF_INET in addresses:
            m_ip = addresses[AF_INET][0]['addr']
            m_ip = netaddr.IPAddress(m_ip).value
            m_network = netaddr.IPNetwork(pon_subnet)
            if m_ip >= m_network.first and m_ip <= m_network.last:
                return iface
    return None


def _get_my_primary_local_ipv4(ifname=None):
    try:
        ifname = get_my_primary_interface() if ifname is None else ifname
        addresses = ni.ifaddresses(ifname)
        ipv4 = addresses[AF_INET][0]['addr']
        return ipv4
    except Exception as e:
        return None


def mac_str_to_tuple(mac):
    return tuple(int(d, 16) for d in mac.split(':'))


if __name__ == '__main__':
    print(get_my_primary_local_ipv4())
